#include <sys/types.h>
#include <sys/socket.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <unistd.h>
#include <string.h>
#include <dlfcn.h>
#include <errno.h>
#include <sys/types.h>
#include <dirent.h>

#include "libfleximon.h"

#define MAX_RESPONSES 1024

static int n_responses = 0;
static int max_responses = MAX_RESPONSES;
struct fm_responder_t responses [MAX_RESPONSES];

static int responses_initialized = 0;
static void init_responses ();

/*********************************************************/
/*                  Logging Functions                    */
/*********************************************************/

/* #define log printf */

static FILE *logfile = NULL;

static void log (char *fmt, ...)
{
  va_list ap;
  if (!logfile) {
    logfile = fopen("fleximon.log","wt");
  }
  va_start(ap,fmt);
  vfprintf(logfile,fmt,ap);
  va_end(ap);
  fflush(logfile);
}


/*********************************************************/
/*            Builtin Response Functions                 */
/*********************************************************/

static char *respond_hello (char *query)
{
  return "Re: HELLO";
}

static char *respond_list (char *query)
{
  /* TODO: replace strcat() with strncat() for security reasons */
  static char buffer [10240];
  int i;
  buffer[0] = 0;
  for (i=0; i<n_responses; i++) {
    if (i) strcat(buffer,",");
    strcat(buffer,responses[i].query);
  }
  return buffer;
}

static char *respond_describe (char *query)
{
  int i;
  log("\tdescribe command: \"%s\"\n",query);
  for (i=0; i<n_responses; i++) {
    if (strcasecmp(query,responses[i].query) == 0)
      return responses[i].description;
  }
  return "unknown command";
}

static char *respond_quit (char *query)
{
  return NULL;
}


/*********************************************************/
/*        Response Function Control Structure            */
/*********************************************************/

void fm_load_builtin (char *query, char *description, char *(*resp)(char*))
{
  if (!responses_initialized) {
    init_responses();
  }
  if (n_responses < max_responses) {
    log("[static] %-15s - %s\n",query,description);
    responses[n_responses].query = query;
    responses[n_responses].description = description;
    responses[n_responses].responder = resp;
    n_responses++;
  } else
    log("[error]  %-15s - cannot register static: too many functions.\n",query);
}

void fm_load_plugin (char *plugin)
{
  void *handle;
  void (*info) (struct fm_responder_t *);
  if (!responses_initialized) {
    init_responses();
  }
  if (n_responses < max_responses) {
    /* log("[plugin] %-15s - loading... \n",plugin); */
    handle = dlopen(plugin,RTLD_LAZY);
    if (!handle) {
      log("[error]  %-15s - cannot load plugin: %s\n",plugin,dlerror());
      return;
    }
    info = dlsym(handle,"plugin_info");
    if (!info) {
      log("[error]  %-15s - plugin info lookup failed: %s\n",plugin,dlerror());
      dlclose(handle);
      return;
    }
    info(&responses[n_responses]);
    log("[dynam.] %-15s - %s\n",responses[n_responses].query,
	responses[n_responses].description);
    n_responses++;
  } else
    log("[error]  %-15s - cannot register plugin: too many functions.\n",
	plugin);
}


void fm_load_plugins (char *plugin_dir)
{
  DIR *pd;
  int namelen;
  char plugname [1024];
  struct dirent *plug;

  pd = opendir(plugin_dir);
  if (!pd) {
    log("[error]  %-15s - cannot open plugin directory: %s\n",
	plugin_dir,strerror(errno));
    return;
  }
  while ((plug = readdir(pd)) != NULL) {
    namelen = strlen(plug->d_name);
    if (namelen < 3) {
      /* not a plugin: filename too short */
      continue;
    }
    if ((plug->d_name[namelen-1] == 'o') &&
	(plug->d_name[namelen-2] == '.')) {
	log("[plugin] %-15s found.\n",plug->d_name);
	strncpy(plugname,plugin_dir,1023);
	strncat(plugname,"/",1023-strlen(plugname));
	strncat(plugname,plug->d_name,1023-strlen(plugname));
	fm_load_plugin(plugname);
    } else {
	/* not a plugin: wrong file extension */
    }
  }
  closedir(pd);
}


/*********************************************************/
/*              Response Control Function                */
/*********************************************************/

char *fm_response (char *query)
{
    int i;
    char *q, *args;
    if (!responses_initialized) {
	init_responses();
    }
    q = query;
    args = strchr(query,' ');
    if (args) {
	*args = 0;
	args++;
    }
    for (i=0; i<n_responses; i++) {
	if (strcasecmp(q,responses[i].query) == 0) {
	    log("[respto] %-15s\n",q);
	    return responses[i].responder(args);
	}
    }
    log("[respto] %-15s - is unknown.\n",query);
    return "UNKNOWN";
}

static void init_responses ()
{
  responses_initialized = 1;
  errno = 0;
  fm_load_builtin("QUIT", "Terminate fleximon daemon", respond_quit);
  fm_load_builtin("HELLO", "Exchange greetings", respond_hello);
  fm_load_builtin("LIST", "List of queries", respond_list);
  fm_load_builtin("DESCRIBE", "Description for query", respond_describe);
  errno = 0;
};


/*********************************************************/
/*              Main Protocol Functions                  */
/*********************************************************/

int fm_run_server (int port, int forwarding)
{
  int server_sockfd, client_sockfd;
  int server_len, client_len;
  struct sockaddr_in server_address;
  struct sockaddr_in client_address;
  char query [1024];
  char resp_buf [2048];
  char *resp;
  int n;

  if (!responses_initialized) {
    init_responses();
  }

  server_sockfd = socket(AF_INET, SOCK_STREAM, 0);
  log("[server] socket() code %2X: %s\n",errno,strerror(errno));

  server_address.sin_family = AF_INET;
  server_address.sin_addr.s_addr = htonl(INADDR_ANY);
  server_address.sin_port = htons(port);

  server_len = sizeof(server_address);
  bind(server_sockfd, (struct sockaddr *)&server_address, server_len);
  log("[server] bind()   code %2X: %s\n",errno,strerror(errno));

  listen(server_sockfd, 5);
  log("[server] listen() code %2X: %s\n",errno,strerror(errno));

  while (1) {
    log("[server] accept()ing.\n");
    client_sockfd = accept(server_sockfd, (struct sockaddr *)&client_address,
			   &client_len);
    log("[server] accept() code %2X: %s\n",errno,strerror(errno));
    n = read(client_sockfd, query, 1023);
    query[n] = 0;
    /* for (i=0; i<n; i++) query[i] = toupper(query[i]); */
    if (forwarding) {
	in_addr_t dest_address;
	unsigned int dest_port;
	char query_bak [1024];
	char *rest;
	strcpy(query_bak,query);
	rest = strchr(query,' ');
	if (rest) *rest++ = 0;
	dest_address = inet_addr(query);
	if (dest_address != INADDR_NONE) {
	    dest_port = atoi(rest);
	    if (dest_port) {
		rest = strchr(rest,' ');
		if (rest) rest++;
	    } else
		dest_port = FM_DEFAULT_PORT;
	    log("[server] forwarding \"%s\" to %X:%i.\n",rest,
		dest_address,dest_port);
	    fm_query(dest_address,dest_port,rest,resp_buf,2048);
	    resp = resp_buf;
	} else {
	    resp = fm_response(query_bak);
	}
    } else {
	resp = fm_response(query);
    }
    if (!resp)
      write(client_sockfd,"QUIT",5);
    else
      write(client_sockfd,resp,strlen(resp)+1);
    close(client_sockfd);
    log("[server] close()  code %2X: %s\n",errno,strerror(errno));
    if (!resp) break;
  }
  log("[server] shutdown()\n");
  shutdown(server_sockfd,2);
  log("[server] shutdown() code %2X: %s\n",errno,strerror(errno));
  close(server_sockfd);
  log("[server] close()  code %2X: %s\n",errno,strerror(errno));
  log("[server] is down.\n");
  return 0;
}


void fm_query (unsigned long host, unsigned port, char *query,
	       char *answer, int answer_maxlen)
{
  int sockfd, len, result;
  struct sockaddr_in address;

  *answer = 0;
  answer[answer_maxlen-1] = 0;
  sockfd = socket(AF_INET, SOCK_STREAM, 0);
  address.sin_family = AF_INET;
  address.sin_addr.s_addr = host;
  address.sin_port = htons(port);
  len = sizeof(address);
  log("[client] connect()ing to %8X:%u\n",host,port);
  result = connect(sockfd, (struct sockaddr *)&address, len);
  if (result == -1) {
    log("[client] connect() code %2X: %s\n",errno,strerror(errno));
    return;
  }
  log("[client] write()  %s\n",query);
  write(sockfd,query,strlen(query)+1);
  len = read(sockfd,answer,answer_maxlen-1);
  answer[len] = 0;
  log("[client] read()   %s\n",answer);
  close(sockfd);
}
