#ifndef LIBFLEXMON_H
#define LIBFLEXMON_H


/*************************************************************
 *                                                           *
 *                         FlexMon                           *
 * Flexible, plugin-based client-server Monitoring framework *
 *                                                           *
 * 2001, 2002, Patrick Schemitz                              *
 *                    <schemitz@ekp.physik.uni-karlsruhe.de> *
 ************************************************************/


/*
    FM_DEFAULT_PORT constant

    the default port number the fleximon server listens to.
 */
#define FM_DEFAULT_PORT 9000


/*
    fm_responder_t structure

    query - the query, e.g. "HELLO" or "LOAD".
    description - a brief description, e.g. "Show system load."
    responder - the responder function.

    Example for a responder function:

    char *respond_sysload (char *arg) { ... }

*/
struct fm_responder_t {
  char *query;
  char *description;
  char *(*responder) (char *);
};


/*
    fm_load_builtin() - register built-in query responder. A builtin
    responder is just a function that takes a char* --the query-- as
    argument and returns a char* --the answer.

    query - the query to which the function responds, e.g. "HELLO".
    description - a few words of description.
    resp - pointer to the responder function.

*/
void fm_load_builtin (char *query, char *description, char *(*resp)(char*));


/*
    fm_load_plugin() - load a plugin query responder. A plugin is
    just a .o file compiled with -DFM_AS_PLUGIN -fPIC -shared, and which
    fills out an FM_PLUGIN() macro.

    plugin - full filename of the plugin.

*/
void fm_load_plugin (char *plugin);
void fm_load_plugins (char *plugin_dir);


/*
    fm_response() - directly respond to a query. Does not use the
    socket-based client-server architecture, i.e. answers within the
    same process. Useful when developing plugins.

    query - the query string.

*/
char *fm_response (char *query);


/*
    fm_run_server() - run FlexMon server on a given port.
    Does not return until a client sends the QUIT command.

    port - the port on which the FlexMon server should listen.
    forwarding - if !=0 enable forwarding (proxy capabilities).

*/
int fm_run_server (int port, int forwarding);


/*
    fm_query() - query a FlexMon server on a given host:port.

    host - the host to be queried, as returned by inet_addr(3).
    port - the port number on which the FlexMon daemon is listening.
    query - the query string.
    answer - the answer is stored here.
    answer_maxlen - maximum answer length.

    Example:

    char answer[1024];
    fm_query(inet_addr("127.0.0.1"),9000,"HELLO",answer,2048);

*/
void fm_query (unsigned long host, unsigned port, char *query,
	       char *answer, int answer_maxlen);


/*
    FM_PLUGIN - macro for plugin registration.

    In plugin mode (-DFM_AS_PLUGIN): generate plugin_info() hook function.
    In test mode: generate a main() function for testing purposes.

    query - name of the query.
    description - brief description what the query does.
    function - responder function.

*/

#ifdef FM_AS_PLUGIN

#define FM_PLUGIN(Query,Description,Function) \
  \
  void plugin_info (struct fm_responder_t *info) \
  { \
    info->query = Query; \
    info->description = Description; \
    info->responder = Function; \
  }

#else

#define FM_PLUGIN(Query,Description,Function) \
  \
  int main (int argc, char **argv) \
  { \
    char *answer; \
    fm_load_builtin(Query,Description,Function); \
    switch(argc) { \
    case 1: \
      printf("Querying list of available queries.\n"); \
      answer = fm_response("LIST"); \
      break; \
    case 2: \
      printf("Query:  \"%s\"\n",argv[1]); \
      answer = fm_response(argv[1]); \
      break; \
    default: \
      fprintf(stderr,"synopsis: %s [query]\n",argv[0]); \
      return 1; \
    } \
    printf("Answer: \"%s\"\n",answer); \
    return 0; \
  }

#endif

#endif
